//  This file is part of jdl_parser.
//
//  jdl_parser is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  jdl_parser is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with jdl_parser.  If not, see <http://www.gnu.org/licenses/>.

// Copyright Simone Pellegrini 2007-2008; e-mail: motonacciu@gmail.com

#include "visitor.hpp"
#include "utils.hpp"
#include "graph_visitor.hpp"

using namespace boost;
using namespace jdl;

///////////////////////////////////////////////
// DUMP VISITOR
///////////////////////////////////////////////
void dump_visitor::visit(int_node const& node){
	//indent(indentation, out);
	out << node.data();
}
void dump_visitor::visit(string_literal_node const& node){
	//indent(indentation, out);
	out << "\"" << node.data() << "\"";
}
void dump_visitor::visit(string_node const& node){
	//indent(indentation, out);
	out << node.data();
}
void dump_visitor::visit(binary_expr_node const& node){
	//indent(indentation, out);
	//indentation++;
	out << "( ";
	node.get_left_operand()->accept(*this);
	out << node.get_operator();
	node.get_right_operand()->accept(*this);
	out << " )";
	//indentation--;
}
void dump_visitor::visit(attribute_definition_node const& node){
	indent(indentation, out);
//	cout << "ATTR_DEF: " << endl;
	//
	node.get_name()->accept(*this);
	out << " = ";
	if(dynamic_pointer_cast<term_node>(node.get_value()) != NULL)
		out << "[" << endl;
	indentation++;
	node.get_value()->accept(*this);
	indentation--;
	if(dynamic_pointer_cast<term_node>(node.get_value()) != NULL)
		out << "]";
	out << ";" << endl;
}
void dump_visitor::visit(conditional_expr const& node){
//	indent(indentation, out);
//	cout << "CONDITIONAL: " << endl;
//	indentation++;
	out << "(";
	node.get_condition()->accept(*this);
	out << ") ? (";
	node.get_then()->accept(*this);
	out << ") : (";
	node.get_else()->accept(*this);
	out << ")";
//	indentation--;
}
void dump_visitor::visit(dot_node const& node){
//	indent(indentation, out);
	node.get_left()->accept(*this);
	out << ".";
	node.get_right()->accept(*this);
}
void dump_visitor::visit(bracket_node const& node){
//	indent(indentation,out);
//	cout << "BRACKET_NODE: " << endl;
//	indentation++;
	node.get_left()->accept(*this);
	out << "[";
	node.get_right()->accept(*this);
	out << "]";
//	indentation--;
}
void dump_visitor::visit(term_node const& node){
//	indent(indentation, out);
//	cout << "TERM: " << endl;
//	indentation++;
	vector<shared_ptr<component> >::const_iterator it = node.get_children().begin();
	while(it != node.get_children().end()){
		(*it++)->accept(*this);
	}
//	indentation--;
}
void dump_visitor::visit(list_node const& node){
//	indent(indentation, out);
//	cout << "TERM: " << endl;
//	indentation++;
	out << "{ ";
	vector<shared_ptr<component> >::const_iterator it = node.get_children().begin();
	while(it != node.get_children().end()){
		(*it++)->accept(*this);
		if(it != node.get_children().end())
			out << ", ";
	}
	out << " }";
//	indentation--;
}
void dump_visitor::visit(func_call_node const& node){
//	indent(indentation, out);
//	cout << "FUNC_CALL: " << endl;
//	indentation++;
	node.get_func_name()->accept(*this);
	out << "(";
	shared_ptr<term_node> term = dynamic_pointer_cast<term_node>(node.get_args());
	assert(term != NULL);
	vector<shared_ptr<component> >::const_iterator it = term->get_children().begin();
	while(it != term->get_children().end()){
		(*it++)->accept(*this);
		if(it != term->get_children().end())
			out << ", ";
	}
//	node.get_args()->accept(*this);
	out << ")";
//	indentation--;
}
void dump_visitor::visit(unary_expr_node const& node){
//	indent(indentation, out);
//	cout << "UNARY_OP: " << node.get_operator() << endl;
//	indentation++;
	out << node.get_operator();
	out << "(";
	node.get_operand()->accept(*this);
	out << ")";
//	indentation--;
}
void dump_visitor::visit(float_node const& node){
//	indent(indentation, out);
	out << node.data();
}

/////////////////////////////////////////////////////
// BASE VISITOR
/////////////////////////////////////////////////////
void base_const_visitor::visit_node(abstract_node const* node){
//	cout << "VISIT: " << typeid(*node).name() << " size: " << node->get_children().size() << endl;
	vector<shared_ptr<component> >::const_iterator it = node->get_children().begin();
	while(it != node->get_children().end())
		(*(it++))->accept(*this);
}

void base_visitor::visit_node(abstract_node *node){
//	cout << "VISIT: " << typeid(*node).name() << " size: " << node->get_children().size() << endl;
	vector<shared_ptr<component> >::iterator it = node->get_children().begin();
	while(it != node->get_children().end())
		(*(it++))->accept(*this);
}

/////////////////////////////////////////////////////
// CHECK TYPE VISITOR
/////////////////////////////////////////////////////
void check_type_visitor::visit(attribute_definition_node const& node){
	string tmp = dynamic_pointer_cast<const string_node>(node.get_name())->data();
	int (*pf)(int)=tolower; 
	transform(tmp.begin(), tmp.end(), tmp.begin(), pf);
	if(tmp == "type"){
		if(found)
			throw semantic_error("Double declaration of attribute Type");
		if(dynamic_pointer_cast<string_node>(node.get_value()) != NULL || dynamic_pointer_cast<string_literal_node>(node.get_value()) != NULL){
			type = dynamic_pointer_cast<const abstract_leaf<string> >(node.get_value())->data();
			transform(type.begin(), type.end(), type.begin(), pf); // to lower case
			found = true;
		}else{
			throw semantic_error("Expected 'string_literal' as Type");
		}
	}
}

/////////////////////////////////////////////////////
// COLLECTION VISITOR
/////////////////////////////////////////////////////
void collection_visitor::visit(attribute_definition_node const& node){
	string tmp = node.get_name()->data();
	int (*pf)(int)=tolower; 
	transform(tmp.begin(), tmp.end(), tmp.begin(), pf);
	
	if(tmp == "nodes"){
	#ifdef DEBUG
		Logger::getInstance().debug().out << "FOUND Collection!" << endl;
	#endif
		assert (dynamic_pointer_cast<list_node>(node.get_value()) != NULL);
		boost::shared_ptr<list_node> children = dynamic_pointer_cast<list_node>(node.get_value());
		vector<boost::shared_ptr<component> >::const_iterator it = children->get_children().begin();
		int i = 0;
		for(;it != children->get_children().end(); it++, i++)
			root->add_child(boost::shared_ptr<graph::node>(new graph::node("node" + boost::lexical_cast<string>(i), (*it).get())));
	}else
		base_const_visitor::visit(node);
}

/////////////////////////////////////////////////////
// DAG VISITOR
/////////////////////////////////////////////////////
void dag_visitor::visit(attribute_definition_node const& node){
	string tmp = node.get_name()->data();
	int (*pf)(int)=tolower; 
	transform(tmp.begin(), tmp.end(), tmp.begin(), pf);
	
	if(tmp == "nodes"){
	#ifdef DEBUG
		Logger::getInstance().debug().out << "FOUND DAG!" << endl;
	#endif
		start_record = true;
		base_const_visitor::visit(node);
		start_record = false;
		found_nodes = true;
		check_dag();
	}else if(start_record && !start_node){
		if(tmp == "dependencies"){
			start_dependecies = true;
			base_const_visitor::visit(node);
			start_dependecies = false;
			found_deps = true;
		}else{
			start_node = true;
			base_const_visitor::visit(node);
			start_node = false;
			if(is_dag_node){
			#ifdef DEBUG
				cout << "FOUND DAG NODE: " << node.get_name()->data() << endl;	
			#endif
				string name = node.get_name()->data();
				if(nodes[name] != NULL)
					throw semantic_error("Node '" + name + "' already defined!");
				
				nodes[name] = node.get_value().get();
				is_dag_node = false;
			}
		}
	}else if(tmp == "dependencies"){
		start_dependecies = true;
		base_const_visitor::visit(node);
		start_dependecies = false;
		found_deps = true;
		check_dag();
	}else if(start_record && start_node){
		if(tmp == "file" || tmp == "description")
			is_dag_node = true;
	}else
		base_const_visitor::visit(node);
}

void build_dependencies(list_node const* node, vector<pair<string,string> >* rules, vector<string>* lit = new vector<string>(), int level = 0){
//	cout << "Building DEPENDENCIES " << node.use_count() << endl;
		
	bool is_complex = false;
//	if(node->get_children().size() > 2)
//		is_complex = false;
	vector<string> s;
	bool first = true;
	vector<shared_ptr<component> >::const_iterator it = node->get_children().begin();
	while(it != node->get_children().end()){
		if(dynamic_pointer_cast<string_node>(*it)){
			string name = dynamic_pointer_cast<string_node>(*it)->data();
			if(first){
			#ifdef DEBUG
				Logger::getInstance().debug().out << "Pushing " << name << endl;
			#endif
				s.push_back(name);
				first = false;
			}else if(!is_complex){
			#ifdef DEBUG
				Logger::getInstance().debug().out << "Pushing " << name << endl;
			#endif
				s.push_back(name);
			}else{
				vector<string>::iterator sit = s.begin();
				while(sit != s.end()){
				#ifdef DEBUG
					Logger::getInstance().debug().out << "Pushing Rule: [" << *sit << ", " << name << "]" << endl;
				#endif
					rules->push_back(pair<string,string>(*sit, name));
					sit++;
				}
				lit->push_back(name);
			}
		}else if(dynamic_cast<list_node *>(it->get()) != NULL){
			is_complex = true;
			assert (node->get_children().size() == 2);
			if(first){
				build_dependencies(dynamic_cast<list_node *>(it->get()), rules, &s, level+1);
				first = false;	
			}else{
				vector<string> right;
				build_dependencies(dynamic_cast<list_node *>(it->get()), rules, &right, level+1);

				vector<string>::iterator sit = s.begin();
				while(sit != s.end()){
					vector<string>::iterator rit = right.begin();
					while(rit != right.end()){
					#ifdef DEBUG
						Logger::getInstance().debug().out << "Pushing Rule: [" << *sit << ", " << *rit << "]" << endl;
					#endif
						rules->push_back(pair<string,string>(*sit,*rit));
						if(sit + 1 == s.end())
							lit->push_back(*rit);
						rit++;
					}
					sit++;
				}
			}
		}
		it++;
	}
	if(!is_complex && level == 0){
		assert (s.size() == 2);
		#ifdef DEBUG
		Logger::getInstance().debug().out << "Pushing Rule: [" << s[0] << ", " << s[1] << "]" << endl;
		#endif
		rules->push_back(pair<string,string>(s[0],s[1]));
	}else{
		vector<string>::iterator sit = s.begin();
		while(sit != s.end())
			lit->push_back(*(sit++));
	}
	if(level == 0)
		delete lit;
}

void throw_node_not_exist(string node){
	throw semantic_error(string("Error occurred checking nodes dependencies; Node '" + node + "' doesn't exist."));
}

void dag_visitor::check_dag(){
	// checking if nodes exist
	if(!found_nodes || !found_deps)
		return;
	vector<pair<string, string> >::iterator it = dependencies.begin();
	while(it != dependencies.end()){
		if(nodes[(*it).first] == NULL)
			throw_node_not_exist((*it).first);
		if(nodes[(*it).second] == NULL)
			throw_node_not_exist((*it).second);
		it++;
	}
	map<string, shared_ptr<graph::node> > dag;
	// BUILDING DAG
	root = new graph::node("ROOT", NULL );
	map<string, component*>::iterator it2 = nodes.begin();
	while(it2 != nodes.end()){
		shared_ptr<graph::node> n = dag[it2->first];
		if(n == 0){
			n = shared_ptr<graph::node>(new graph::node(it2->first, it2->second));
			dag[it2->first] = n;
		}
		bool is_root = true;
		vector<pair<string, string> >::iterator it3 = dependencies.begin();
		while(it3 != dependencies.end()){
			if(it3->first == it2->first){
				shared_ptr<graph::node> child = dag[it3->second];
				if(child == 0){
					child = shared_ptr<graph::node>(new graph::node(it3->second, nodes[it3->second]));
					dag[it3->second] = child;
				}
				n->add_child(child);
			}else if(it3->second == it2->first)
				is_root = false;
			it3++;
		}
		if(is_root)
			root->add_child(n);
		it2++;
	}
	// CHECKing A-CYCLIC
	graph::acyclic_visitor cv;
	root->accept(cv);
}

void dag_visitor::visit(list_node const& node){
	if(start_dependecies && !dep_list){
	#ifdef DEBUG
		Logger::getInstance().debug().out << "FOUND DEPENDECIES" << endl;
	#endif
		dep_list = true;
		base_const_visitor::visit(node);
		dep_list = false;
	}else if(start_dependecies && dep_list){
		build_dependencies(&node, &dependencies);
	}else
		base_const_visitor::visit(node);
}

/////////////////////////////////////////////////////////////////////////
// Find Visitor
/////////////////////////////////////////////////////////////////////////
find_visitor::find_visitor(std::string const& _path): p(_path), depth(0), value(NULL), attribute(NULL){
	string::const_iterator it = _path.begin();
//	if(*it != '.')
//		throw logic_error("Bad path format. Path must start with '/'.");
//	it++; // skip the '/'
	string::const_iterator start = it;
	while(it != _path.end()){
		if(*it == '.'){
			string tmp = string(start,it);
			int (*pf)(int)=tolower; 
			transform(tmp.begin(), tmp.end(), tmp.begin(), pf);
			if(tmp != "root"){
				path.push_back(tmp);
				selector.push_back(-1);
			}
			start = it + 1;
		}else if(*it == '['){
			string tmp = string(start,it);
			int (*pf)(int)=tolower; 
			transform(tmp.begin(), tmp.end(), tmp.begin(), pf);
			path.push_back(tmp);
			it++; // skip the '['
			start = it;
			string::const_iterator fit = find(it, _path.end(), ']');
			selector.push_back(lexical_cast<int>(string(start,fit)));
			it = fit;
			it++; // skip the ']'
			start = it + 1; // skip the '.'
			if(it == _path.end())
				return;
		}
		it++;
	}
	if(start != it){
		// last node
		string tmp = string(start,it);
		int (*pf)(int)=tolower; 
		transform(tmp.begin(), tmp.end(), tmp.begin(), pf);
		path.push_back(tmp);
		selector.push_back(-1);
	}

//	for(vector<string>::const_iterator it = path.begin(); it != path.end(); it++)
//		cout << *it << " ";
//	cout << endl;
//	for(vector<int>::const_iterator it = selector.begin(); it != selector.end(); it++)
//		cout << *it << " ";
//	cout << endl;
}

void find_visitor::visit(attribute_definition_node const& node){
	//cout << "DEPTH : " << depth << " PATH: " << path[depth] <<  endl;
	//cout << "NODE NAME: " << node.get_name()->data() << endl;
	if(value != NULL)
		return;
	
	string tmp_name = node.get_name()->data();
	int (*pf)(int)=tolower; 
	transform(tmp_name.begin(), tmp_name.end(), tmp_name.begin(), pf);
	
	component* n = node.get_value().get();
	
	if(tmp_name == "description") // skip the description node
		n->accept(*this);
	
	if(path[depth] == tmp_name){
		if(selector[depth] != -1){
			if(dynamic_pointer_cast<abstract_node>(node.get_value()) != NULL)
				n = dynamic_pointer_cast<abstract_node>(node.get_value())->get_children()[selector[depth]].get();
			else
				throw logic_error("Error while looking for path: " + p + ", node '" + tmp_name + "' doesn't contains children nodes.");
		}
		if(depth == path.size()-1){
			value = n;
			attribute = &node;
			return;
		}
		
		depth++;
		n->accept(*this);
		depth--;
	}
}
